/* serverpacket.c
 *
 * Constuct and send DHCP server packets
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <time.h>

#include "packet.h"
#include "debug.h"
#include "dhcpd.h"
#include "options.h"
#include "leases.h"
#include "files.h"

// Added by Mason Yu for Half Bridge
#include <net/if.h>
#include <linux/sockios.h>
#if defined(IP_BASED_CLIENT_TYPE)
#include <stdio.h>
#include <stdlib.h>
#endif

#ifdef CTC_DHCP_OPTION43
#define CONFIGVERSTR0     'C'
#define CONFIGVERSTR1     'T'
#define CONFIGVERSTR2     'C'
#define CONFIGVERSTR3     '0'
char option43[1+1+254]=
	{43, 6, 1, 4, CONFIGVERSTR0, CONFIGVERSTR1, CONFIGVERSTR2, CONFIGVERSTR3,0,0};
#endif
#ifdef CTC_DHCP_OPTION60
char option60[1+1+254]=
	{60, 2, 0, 0,}; // FOR dhcp ack, we just send enterprise code to clients.
const char IPTABLES[] = "/bin/iptables";
#endif

#if defined(IP_BASED_CLIENT_TYPE)
static char *g_apszCTCDeviceName[] = 
{
    "Computer",
    "Camera",
    "HGW",
    "STB",
    "Phone",
    NULL
};
#endif

#ifdef _CWMP_TR069_

//#define _TR111_TEST_
#define TR111_DEVICEFILE	"/tmp/tr111device.txt"

int handle_tr111_save( char *buf, int *ins_num )
{
	FILE *fp;
	char tmp[160];
	int found=0;
	
	if( buf==NULL || ins_num==NULL ) return -1;
	
	fp = fopen( TR111_DEVICEFILE, "r" );
	if(fp)
	{
		char *p;
		while( fgets(tmp,160,fp) )
		{
			p = strtok( tmp, " \r\n" );
			if(p)
			{
				p = strtok( NULL, "\r\n" );
				if(  p!=NULL && strcmp(p,buf)==0 )
					found = 1;
			}
		}
		fclose(fp);
	}
	
	if(found==0)
	{
		fp=fopen( TR111_DEVICEFILE, "a" );
		if(fp)
		{
			sprintf( tmp,"%d %s\n", *ins_num,buf );
			fwrite( tmp, 1, strlen(tmp), fp );
			fclose(fp);
			(*ins_num)++;
		}
	}
	return 0;
}


int handle_tr111(struct dhcpMessage *oldpacket)
{
	int ret=0;
	static int ins_num=1;
	unsigned char *pOpt125=NULL;
#ifndef _TR111_TEST_
	pOpt125 = get_option(oldpacket, DHCP_VI_VENSPEC);
#else
      #if 0
	unsigned char testbuf[]={ 0x00, 0x00, 0x0d, 0xe9, 0x18, 0x01, 0x06, 0x30,
				  0x30, 0x65, 0x30, 0x64, 0x34, 0x02, 0x09, 0x30,
				  0x30, 0x30, 0x30, 0x2d, 0x30, 0x30, 0x30, 0x31,
				  0x03, 0x03, 0x49, 0x47, 0x44, 0x00};
      #else
	unsigned char testbuf[]={
	0x00 , 0x00 , 0x0d , 0xe9 , 0x2d , 0x01 , 0x06 , 0x30 , 0x30 , 0x30,
	0x31 , 0x30 , 0x32 , 0x02 , 0x10 , 0x30 , 0x30 , 0x30 , 0x31 , 0x30 , 0x32 , 0x2d , 0x34 , 0x32 , 0x38 , 0x32,
	0x38 , 0x38 , 0x38 , 0x32 , 0x39 , 0x03 , 0x11 , 0x43 , 0x44 , 0x52 , 0x6f , 0x75 , 0x74 , 0x65 , 0x72 , 0x20,
	0x56 , 0x6f , 0x49 , 0x50 , 0x20 , 0x41 , 0x54 , 0x41 };
      #endif
	pOpt125 = testbuf;
#endif


	if(pOpt125)
	{
		unsigned int ent_num, *pUInt;
		unsigned short data_len;
		char GW_OUI[7]="",GW_SN[65]="",GW_CLASS[65]="";
		char buf[160];
		
		pUInt = (unsigned int*)pOpt125;
		ent_num = ntohl( *pUInt );
		data_len = (unsigned short)pOpt125[4];

		//if more than one enterprise-number in the same option125??
		if( ent_num==3561 )
		{
			unsigned char *pStart;
			
			//sub-option
			pStart = &pOpt125[5];
			while( data_len>0 )
			{
				unsigned char sub_code, sub_len, *sub_data;
				
				sub_code = pStart[0];
				sub_len = pStart[1];
				sub_data = &pStart[2];
				
				if( data_len < sub_len+2 )
					break;
					
				switch( sub_code )
				{
				case 1://DeviceManufacturerOUI
					if( sub_len<7 )
					{
						strncpy( GW_OUI, sub_data, sub_len );
						GW_OUI[sub_len]=0;
					}
					break;
				case 2://DeviceSerialNumber
					if( sub_len<65 )
					{
						strncpy( GW_SN, sub_data, sub_len );
						GW_SN[sub_len]=0;
					}
					break;
				case 3://DeviceProductClass
					if( sub_len<65 )
					{
						strncpy( GW_CLASS, sub_data, sub_len );
						GW_CLASS[sub_len]=0;
					}
					break;
				default:
					//unknown suboption
					break;
				}
				
				
				pStart = pStart+2+sub_len;
				data_len = data_len-sub_len-2;
			}

			
			LOG(LOG_ERR,  "%d: oui:%s sn:%s class:%s ", __LINE__,GW_OUI,GW_SN,GW_CLASS  );
			if( *GW_OUI && *GW_SN && *GW_CLASS )
			{
				sprintf( buf, "%s?%s?%s", GW_OUI,GW_CLASS,GW_SN );
				handle_tr111_save( buf, &ins_num );
			}else if( *GW_OUI && *GW_SN  )
			{
				sprintf( buf, "%s?%s", GW_OUI,GW_SN );
				handle_tr111_save( buf, &ins_num );
			}else{
				//error
			}			
		}
	}
	
	return ret;
}
#endif

#if defined(IP_BASED_CLIENT_TYPE)
int parse_CTC_Vendor_Class(struct dhcpMessage *packet, unsigned char* option60, struct dhcp_ctc_client_info *pstClientInfo)
{
	unsigned short enterprise_code=(*option60<<8)+*(option60+1);
	unsigned char *fieldtype;
	unsigned char fieldlength;
	int iOptlen, i;
    
	DEBUG_CHN_TEL("__FUNCTION__, enterprise_code=%04x\n", enterprise_code);

	pstClientInfo->iCategory = CTC_Computer;
    
	if(enterprise_code!=0x0000)
	{	
		return -1;// not china telecom enterprise code.. 
	}
    
	fieldtype=(option60+2);
    iOptlen = *(option60-1);
    while (iOptlen > 0)
    {
        fieldlength = *(fieldtype + 1);
        switch (*fieldtype)
        {
            case Vendor:
                if ((fieldlength < DHCP_CTC_MIN_FIELD_LEN)
                    || (fieldlength > DHCP_CTC_MAX_FIELD_LEN))
                {
                    return -1;
                }

                memcpy(pstClientInfo->szVendor, fieldtype + 2, fieldlength);
                pstClientInfo->szVendor[fieldlength] = 0;
                break;

            case Category:
                if ((fieldlength < DHCP_CTC_MIN_FIELD_LEN)
                    || (fieldlength > DHCP_CTC_MAX_FIELD_LEN))
                {
                    return -1;
                }
                
                for (i = 0; g_apszCTCDeviceName[i]; i++)
                {
                    if ((fieldlength == strlen(g_apszCTCDeviceName[i]))
                        && (0 == memcmp(fieldtype + 2, g_apszCTCDeviceName[i], fieldlength)))
                    {
                        pstClientInfo->iCategory = i;
                        break;
                    }
                }
                break;

            case Model:
                if ((fieldlength < DHCP_CTC_MIN_FIELD_LEN)
                    || (fieldlength > DHCP_CTC_MAX_FIELD_LEN))
                {
                    return -1;
                }

                memcpy(pstClientInfo->szModel, fieldtype + 2, fieldlength);
                pstClientInfo->szModel[fieldlength] = 0;
                break;

            case Version:
                if ((fieldlength < DHCP_CTC_MIN_FIELD_LEN)
                    || (fieldlength > DHCP_CTC_MAX_FIELD_LEN))
                {
                    return -1;
                }

                memcpy(pstClientInfo->szVersion, fieldtype + 2, fieldlength);
                pstClientInfo->szVersion[fieldlength] = 0;
                break;

            case ProtocolType:
                memcpy((char *)(&pstClientInfo->stPortForwarding.usProtocol),
                        fieldtype + 2,
                        sizeof(unsigned short));
                memcpy((char *)(&pstClientInfo->stPortForwarding.usPort),
                        fieldtype + 4,
                        sizeof(unsigned short));
                break;

            default:
                break;
        }

        iOptlen -= fieldlength + 2;
        fieldtype += fieldlength + 2;
    }

	return 0;
}

int check_type(u_int32_t addr, enum DeviceType devicetype)
{
	int ret=0;

	switch(devicetype)
	{
		case CTC_Computer:
			if(addr<ntohl(server_config.pcstart)||addr>ntohl(server_config.pcend))
				ret=1;
			break;
		case CTC_Camera:
			if(addr<ntohl(server_config.cmrstart)||addr>ntohl(server_config.cmrend))
				ret=1;
			break;
		case CTC_STB:
			if(addr<ntohl(server_config.stbstart)||addr>ntohl(server_config.stbend))
				ret=1;
			break;
		case CTC_PHONE:
			if(addr<ntohl(server_config.phnstart)||addr>ntohl(server_config.phnend))
				ret=1;
			break;
		case CTC_HGW:
			if(addr<ntohl(server_config.hgwstart)||addr>ntohl(server_config.hgwend))
				ret=1;
			break;
		case CTC_UNKNOWN:
			if(addr<ntohl(server_config.start)||addr>ntohl(server_config.end))
				ret=1;
			if((addr>ntohl(server_config.pcstart)&&addr<ntohl(server_config.pcend))||
			   	(addr>ntohl(server_config.cmrstart)&&addr<ntohl(server_config.cmrend))||
				(addr>ntohl(server_config.stbstart)&&addr<ntohl(server_config.stbend))||
				(addr>ntohl(server_config.phnstart)&&addr<ntohl(server_config.phnend)))
				ret=1;
			break;
		default:
			break;
	}
	
	return ret;
}

#endif
//added by jim luo to support china telecom e8 spec.
#ifdef CONFIG_CTC_E8_CLIENT_LIMIT
#if 0
/*
static int acceptedAllowedNumbers(struct CTC_Clients *clients, int range);
static void blockingClient(unsigned int ip)
{
	unsigned char buffer[128];
		//va_cmd("/bin/iptables"  6, 1, "-I", "FORWARD", "-s", ip, "-j", "DROP");
	sprintf(buffer, "/bin/iptables -I FORWARD -s %d.%d.%d.%d -j DROP", 
					(ip>>24)&0xFF, (ip>>16)&0xFF, (ip>>8)&0xFF, (ip>>0)&0xFF);

	system(buffer);
	printf("blocking client %08x\n", ip);
}
static void unblockingClient(unsigned int ip)
{
	unsigned char buffer[128];
	//va_cmd("/bin/iptables"  6, 1, "-D", "FORWARD", "-s", ip, "-j", "DROP");
	sprintf(buffer, "/bin/iptables -D FORWARD -s %d.%d.%d.%d -j DROP", 
					(ip>>24)&0xFF, (ip>>16)&0xFF, (ip>>8)&0xFF, (ip>>0)&0xFF);

	system(buffer);
	printf("unblocking client %08x\n", ip);
}

//push out the unused elements in list.....
static void reorderClientsList(struct CTC_Clients *clients, int range)
{
	int i;
	for(i=0;  i<range; i++)
	{
		if(clients[i].ip==0)
		{	//copress
			int j,k;
			j=i;
			for(k=0; k<range-i-1; k++, j++)
				memcpy(&clients[j], &clients[j+1], sizeof(struct CTC_Clients)); 
		}
	}
}
static void removeUnexistClients(struct CTC_Clients *clients, int range)
{
	int i;
	for(i=0; i< range; i++)
	{
		if((clients[i].ip!=0)&&(clients[i].detected==0))
		{	// the blocked connection is not detected, then unblock it from iptables.
			if(clients[i].state==0) //blocked
				unblockingClient(clients[i].ip); //remove the iptables blocking entry
			clients[i].ip=0;
			clients[i].state=0;
		}
		if(clients[i].ip==0)
		{	//copress
			int j,k;
			j=i;
			for(k=0; k<range-i-1; k++, j++)
				memcpy(&clients[j], &clients[j+1], sizeof(struct CTC_Clients)); 
		}
		
	}
	return ; 
}
static int updateClientsState(struct CTC_Clients *clients, int range, int maxclients)
{
	int clientsAllowedAlready;
	int clientsCanAllowed;
	int clientsAllowedFromBlocking=0;
	int i,j;
	clientsAllowedAlready=acceptedAllowedNumbers(clients, range);
	clientsCanAllowed=maxclients-clientsAllowedAlready;   //we can allow some blocked connections to be allowed to access WAN..
	for(i=0,j=0; i< clientsCanAllowed; i++)
	{
		for( ; j<range; j++)
		{
			if(clients[j].ip && clients[j].state==0)  //blocked state
			{
				clients[j].state=1; // allowed state.
				unblockingClient(clients[j].ip);
				clientsAllowedFromBlocking++;
				break;
			}
		}
	}
	return clientsAllowedAlready+clientsAllowedFromBlocking; // return the updated allow/block connection nunbers to caller.
}
static int clientInList(struct CTC_Clients *clients, int range,  unsigned long ip, int * state)
{
	int i;
	for(i=0; i< range; i++)
	{
		if(clients[i].ip==0)
			return -1;
		if(clients[i].ip==ip)
		{
			*state=clients[i].state;
			//clients[i].detected=1; // set detected flags.
			return i; // found..
		}
	}
	return -1; //not found..
}
static int acceptedAllowedNumbers(struct CTC_Clients *clients, int range)
{
	int i;
	int numbers=0;
	for(i=0; i<range; i++)
	{
		if( ( clients[i].ip!=0) &&(clients[i].state==1)) //unblocking
			numbers++;
	}
	return numbers;
}
static int insertClients(struct CTC_Clients  *clients, int range, unsigned int ip, int state)
{
	int i;
	for(i=0; i<range; i++)
	{
		if(clients[i].ip==ip || clients[i].ip==0)
		{
			clients[i].ip=ip;
			clients[i].state=state;
			return 0;
		}
	}
	return -1;
}
static struct CTC_Clients *getNextClient(struct CTC_Clients * clients, int *range)
{
	int i;
	if(*range==0)
		return NULL;
	for(i=0;i<*range; i++)
	{
		if(clients->ip!=0)
		{
			*range=*range-i-1;
			return clients;
		}
		clients++;
	}
	return NULL;
}


static int checkClients(unsigned int ip, enum DeviceType devtype)
{
	//using Arp request to detect the accepted-clients is still here.
	struct CTC_Clients * client;
	unsigned char buffer[1024];
	unsigned char recv_buffer[1024];
	unsigned char *curptr;
	int len,i;
	unsigned int ipparts[4];
	int range;
	int pcNum=0;
	int cameraNum=0;
	int pcNumAllowed=0;
	int pcNumBlocked=0;
	int cameraNumAllowed=0;
	int cameraNumBlocked=0;
	FILE * fd;
	curptr=buffer;
	//printf("checkClients in\n");
	fd=fopen(CLIENTSMONITOR, "r+");
	if(fd==NULL)
		return -1;
	//printf("checkClients in 2\n");
	memset(buffer, 0, sizeof(buffer));
	// send Arp Request to eavy accepted-clients
	client=accepted_PC_Clients;
	range=MAXPCCLIENTS;
	len=sprintf(curptr, "1 "); // set detect flag...
	if(len < 0)
		goto err;
	curptr+=len;
	
	while(client=getNextClient(client, &range))
	{
		//sendArpRequest(*client);
		len=sprintf(curptr, "%d.%d.%d.%d ", (client->ip>>24)&0xFF, (client->ip>>16)&0xFF,
										(client->ip>>8)&0xFF, (client->ip>>0)&0xFF); // set ip to detect...
		if(len < 0)
			goto err;
		curptr+=len;
		client++; //next client
	}
	//printf("checkClients in 3\n");
	client=accepted_Camera_Clients;
	range=MAXCAMERACLIENTS;
	while(client=getNextClient(client, &range))
	{
		//sendArpRequest(*client);
		len=sprintf(curptr, "%d.%d.%d.%d ", (client->ip>>24)&0xFF, (client->ip>>16)&0xFF,
										(client->ip>>8)&0xFF, (client->ip>>0)&0xFF); // set ip to detect...
		if(len < 0)
			goto err;
		curptr+=len;
		client++;
	}
	//printf("checkClients in 4\n");
	// before gather detected clients info, zero detected flag...
	for(i=0; i< MAXPCCLIENTS; i++)
	{
		accepted_PC_Clients[i].detected=0;
	}
	for(i=0; i< MAXCAMERACLIENTS; i++)
	{
		accepted_Camera_Clients[i].detected=0;
	}


// do 3 times try detect...
{
	int localcount=0;
retry:
	//printf("retry count\n");
	fclose(fd);
	fd=fopen(CLIENTSMONITOR, "r+");
	fwrite(buffer, 1, strlen(buffer),fd);
	fflush(fd);
	//though china telecom spec need 3 times detect to determined if still existed, but one time is enough...
	usleep(200000);// wait arp reply
	memset(recv_buffer, 0, sizeof(recv_buffer));
	curptr=recv_buffer;
	//printf("checkclients: read\n");
	len=fread(curptr, 1,  1024, fd);
	//fscanf(fd, "%-16s 0x%-8x0x%-6x%s"
	//						"   *     %s\n", 
	if(len < 0)
		goto err;
	//printf("checkClients in 5\n");
	printf("%s\n", curptr);

	for(i=0; i< MAXCAMERACLIENTS+MAXPCCLIENTS; i++)
	{
		unsigned long ip;
		int state;
		int index;
		if(sscanf(curptr, "%d.%d.%d.%d", &ipparts[0],&ipparts[1],&ipparts[2],&ipparts[3])<=0)
			break;
		ip=(ipparts[0]<<24) |( ipparts[1]<<16) |( ipparts[2]<<8) |( ipparts[3]<<0);
		if((index=clientInList(accepted_PC_Clients, MAXPCCLIENTS, ip, &state))>=0)
		{
			accepted_PC_Clients[index].detected=1;
		}else if((index=clientInList(accepted_Camera_Clients, MAXCAMERACLIENTS, ip, &state))>=0)
		{
			accepted_Camera_Clients[index].detected=1;
		}
		//search next ip.
		while((*curptr != ' ') && (*curptr != 0))
		{
			curptr++;
		}
		if(*curptr==0)
			break;
		curptr++;
	}
	localcount++;
	if(localcount<3)
		goto retry;
}
	// now all the clients-connected have been polled. 
	removeUnexistClients(accepted_PC_Clients,MAXPCCLIENTS);
	removeUnexistClients(accepted_Camera_Clients, MAXCAMERACLIENTS);
	pcNumAllowed=updateClientsState(accepted_PC_Clients,MAXPCCLIENTS, maxPCClients);
	cameraNumAllowed=updateClientsState(accepted_Camera_Clients,MAXCAMERACLIENTS, maxCameraClients);
	if(devtype == CTC_Carema)
	{
		int state;
		if(cameraNumAllowed<maxCameraClients)
			state=1;   //allowed
		else
		{
			state=0;
			blockingClient(ip);
		}
		insertClients(accepted_Camera_Clients,MAXCAMERACLIENTS, ip, state);
		
		
	}else // Computers device default
	{
		int state;
		if(pcNumAllowed<maxPCClients)
			state=1;   //allowed
		else
		{
			state=0;
			blockingClient(ip);
		}
		insertClients(accepted_PC_Clients,MAXPCCLIENTS, ip, state);
	}

//reset arp-detecting to idle
	memset(buffer, 0, sizeof(buffer));
	curptr=buffer;
	len=sprintf(curptr, "0 "); // set detect flag...
	if(len < 0)
		goto err;
	//printf("checkClients in 6\n");
	fwrite(buffer, 1, strlen(buffer),fd);
	fclose(fd);
	return 0;
err:
	//printf("checkClients err out\n");
	fclose(fd);
	return -1;
	
	
}*/

static void handle_clients_limit(struct dhcpMessage *packet,unsigned long yiaddr)
{
	unsigned char *classVendor;
	enum DeviceType      devicetype;
	char pass_buffer[256];
	char *curptr=pass_buffer;
	int len;
	FILE * fd;
    struct dhcp_ctc_client_info stClientInfo;
	//printf("checkClients in\n");
	
	//check the device type: PC or Camera...
	printf("handle_clients_limit in\n");
	if(!(classVendor=get_option(packet, DHCP_VENDOR)))
		//default : PC clients....
		devicetype=CTC_Computer;
	else
	{
        memset(&stClientInfo, 0, sizeof(struct dhcp_ctc_client_info));
		parse_CTC_Vendor_Class(packet, classVendor, &stClientInfo);
        devicetype = (enum DeviceType)(stClientInfo.iCategory);
	}
	// send it to kernel...
	fd=fopen(CLIENTSMONITOR, "r+");
	if(fd==NULL)
	{
		printf("no such proc file %s\n", CLIENTSMONITOR);
		return;
	}
	memset(pass_buffer, 0, sizeof(pass_buffer));

//debug management path through proc/ClientsMonitor.....
if(0) // test pass, below path is OK....
{
	len=sprintf(curptr, "flagLimitOnAll "); // set ip to detect...
	if(len < 0)
		goto err;
	curptr+=len;
	len=sprintf(curptr, "%d\n", 0);
	curptr+=len;
	
	len=sprintf(curptr, "limitOnAll "); // set ip to detect...
	if(len < 0)
		goto err;
	curptr+=len;
	len=sprintf(curptr, "%d\n", 252);
	curptr+=len;

	len=sprintf(curptr, "computerLimitEnable "); // set ip to detect...
	if(len < 0)
		goto err;
	curptr+=len;
	len=sprintf(curptr, "%d\n", 0);
	curptr+=len;
	
	len=sprintf(curptr, "limitOnComputer "); // set ip to detect...
	if(len < 0)
		goto err;
	curptr+=len;
	len=sprintf(curptr, "%d\n", 201);
	curptr+=len;
	
	len=sprintf(curptr, "cameraLimitEnable "); // set ip to detect...
	if(len < 0)
		goto err;
	curptr+=len;
	len=sprintf(curptr, "%d\n", 1);
	curptr+=len;
	
	len=sprintf(curptr, "limitOnCamera "); // set ip to detect...
	if(len < 0)
		goto err;
	curptr+=len;
	len=sprintf(curptr, "%d\n", 39);
	curptr+=len;
}
	//the format should conformance with kernel arp.c
	len=sprintf(curptr, "IP              Types\n"); // set ip to detect...
	if(len < 0)
		goto err;
	curptr+=len;
	len=sprintf(curptr, "%d.%d.%d.%d %d\n", (yiaddr>>24)&0xFF, (yiaddr>>16)&0xFF,
										(yiaddr>>8)&0xFF, (yiaddr>>0)&0xFF, (int)devicetype); // set ip to detect...
	if(len < 0)
		goto err;
	curptr+=len;
	len=fwrite(pass_buffer, 1, 2+(unsigned)curptr-(unsigned)pass_buffer, fd);	//some trail zero should be included.	
	fclose(fd);
	printf("handle_clients_limit exit len=%d\n", len);	
	if(1)
	{
			printf("proc wirrten: %s", pass_buffer);
	}
	return;
err:
	fclose(fd);
	return;
	
}
#endif
#endif
/* send a packet to giaddr using the kernel ip stack */
static int send_packet_to_relay(struct dhcpMessage *payload)
{
	DEBUG(LOG_INFO, "Forwarding packet to relay");

	return kernel_packet(payload, server_config.server, SERVER_PORT,
			payload->giaddr, SERVER_PORT);
}


/* send a packet to a specific arp address and ip address by creating our own ip packet */
static int send_packet_to_client(struct dhcpMessage *payload, int force_broadcast)
{
	unsigned char *chaddr;
	u_int32_t ciaddr;
	
	if (force_broadcast) {
		DEBUG(LOG_INFO, "broadcasting packet to client (NAK)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else if (payload->ciaddr) {
		DEBUG(LOG_INFO, "unicasting packet to client ciaddr");
		ciaddr = payload->ciaddr;
		chaddr = payload->chaddr;
	} else if (ntohs(payload->flags) & BROADCAST_FLAG) {
		DEBUG(LOG_INFO, "broadcasting packet to client (requested)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else {
		DEBUG(LOG_INFO, "unicasting packet to client yiaddr");
		ciaddr = payload->yiaddr;
		chaddr = payload->chaddr;
	}
	return raw_packet(payload, server_config.server, SERVER_PORT, 
			ciaddr, CLIENT_PORT, chaddr, server_config.ifindex);
}


/* send a dhcp packet, if force broadcast is set, the packet will be broadcast to the client */
static int send_packet(struct dhcpMessage *payload, int force_broadcast)
{
	int ret;

	if (payload->giaddr)
		ret = send_packet_to_relay(payload);
	else ret = send_packet_to_client(payload, force_broadcast);
	return ret;
}


static void init_packet(struct dhcpMessage *packet, struct dhcpMessage *oldpacket, char type)
{
	init_header(packet, type);
	packet->xid = oldpacket->xid;
	memcpy(packet->chaddr, oldpacket->chaddr, 16);
	packet->flags = oldpacket->flags;
	packet->giaddr = oldpacket->giaddr;
	packet->ciaddr = oldpacket->ciaddr;
	add_simple_option(packet->options, DHCP_SERVER_ID, server_config.server);
}


/* add in the bootp options */
static void add_bootp_options(struct dhcpMessage *packet)
{
	packet->siaddr = server_config.siaddr;
	if (server_config.sname)
		strncpy(packet->sname, server_config.sname, sizeof(packet->sname) - 1);
	if (server_config.boot_file)
		strncpy(packet->file, server_config.boot_file, sizeof(packet->file) - 1);
}


// Added by Mason Yu for MAC Base assignment
unsigned long find_IP_by_Mac(unsigned char * packet_chaddr)
{
	char tmp_mac[6][3];
	char chaddr[19];
	FILE *fp;
	char temps[0x100];
	unsigned long value;
	char *str, *endptr;
	struct in_addr matchIp;
	int offset;				
	int i;
	
	for (i=0; i<6; i++) {
		if ( packet_chaddr[i] <= 0xf )
			sprintf(tmp_mac[i], "0%x",  packet_chaddr[i]);
		else
			sprintf(tmp_mac[i], "%x",  packet_chaddr[i]);	        		
	}
	
	sprintf(chaddr, "%s-%s-%s-%s-%s-%s:", tmp_mac[0], tmp_mac[1], tmp_mac[2], tmp_mac[3], tmp_mac[4], tmp_mac[5]);
	//printf("chaddr=%s\n", chaddr);	        		
	
	value = 0;
	
	if ((fp = fopen("/var/dhcpdMacBase.txt", "r")) == NULL)
	{
		printf("Open file /var/dhcpdMacBase.txt fail !\n");
		return 0;
	}				
	
	while (fgets(temps,0x100,fp))
	{
		if (temps[strlen(temps)-1]=='\n')
			temps[strlen(temps)-1] = 0;
		
		if (str=strstr(temps, chaddr))
		{							
			offset = strlen(chaddr)+1;			
			//printf("The string is %s\n", str+offset);
									
			inet_aton(str+offset, &matchIp);	
			value = matchIp.s_addr;	
			//printf("value=0x%x\n", value);						
			break;	
			
		} else {					
			continue;
			
	        }
		
	}				
	
	return value;							        		
}

//star add: for static ip based Mac
int find_Mac_by_IP(u_int32_t ipaddr)
{
	struct in_addr matchIp;
	char ip[20];
	char * str=0;
	FILE *fp;
	int value=0;
	char temps[0x100];

	matchIp.s_addr = ipaddr;
	strcpy(ip,inet_ntoa(matchIp));

	if ((fp = fopen("/var/dhcpdMacBase.txt", "r")) == NULL)
	{
		printf("Open file /var/dhcpdMacBase.txt fail !\n");
		return 0;
	}	

	while (fgets(temps,0x100,fp))
	{
		if (temps[strlen(temps)-1]=='\n')
			temps[strlen(temps)-1] = 0;
		
		if (str=strstr(temps, ip))
		{							
			value = 1;				
			break;	
			
		} else {					
			continue;
		}	
	}		
	//jim: we should release all resource, fixed by jim
	fclose(fp);
	return value;
	
}
	
// Added by Mason Yu for PPP Half Bridge
int PPP_HALFBRIDGE_GET_WANIP = FALSE;
//int PPP_HALFBRIDGE_GET_WANIP = TRUE;
int IPOA_HALFBRIDGE_GET_WANIP = FALSE;
int FIRST_SEND_PUBLIC_IP = FALSE;	
int FIRST_SEND_ACK = TRUE;
int RENEW_PUBLIC_IP = FALSE;

/* send a DHCP OFFER to a DHCP DISCOVER */
int sendOffer(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;
	struct dhcpOfferedAddr *lease = NULL;
	u_int32_t req_align, lease_time_align = server_config.lease;
	unsigned char *req, *lease_time;
	struct option_set *curr;
	struct in_addr addr;

    	// Added by Mason Yu for Half Bridge
    	// Added by Mason Yu. Access internet fail.
    	unsigned long myipaddr, hisipaddr, dnsipaddr;
    	FILE *fp;
    	struct ifreq ifr;
    	int i;       
    	static int skfd = -1;
    	//struct sockaddr_in *addrtemp;
    	//struct in_addr pAddr;
    	//char incoming_route[100];

    	init_packet(&packet, oldpacket, DHCPOFFER);    
    
    	// Added by Mason Yu for Half Bridge, Start
    	//printf("server_config.ippt = %d\n", server_config.ippt);
    	if ( server_config.ippt )  {
        	if ((skfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			perror("creat socket fail");
			return (0);
		}
             
         	// If we can get WAN IP, that is the WAN interface is up.  
         	/*                 
         	if (skfd >= 0) {
             		strcpy(ifr.ifr_name, "ppp0");
             		ifr.ifr_addr.sa_family = AF_INET;
             		if (ioctl(skfd, SIOCGIFADDR, &ifr) == 0) {
             			PPP_HALFBRIDGE_GET_WANIP = TRUE;
            		 }
         	}
         	*/            
         	 
         	// Read global variables from sppp process.
	 	fp = fopen ("/tmp/PPPHalfBridge", "r");
         	if (fp) {
	 	       fread(&myipaddr, 4, 1, fp);	 	         
	 	       fread(&hisipaddr, 4, 1, fp);
	 	       // Added by Mason Yu. Access internet fail.
	 	       fread(&dnsipaddr, 4, 1, fp);        	   
	 	       fclose(fp);
	 	       PPP_HALFBRIDGE_GET_WANIP = TRUE;
         	} else {
         		PPP_HALFBRIDGE_GET_WANIP = FALSE;
         	}   
         
         	// Read global variables from goahead process.
	 	fp = fopen ("/tmp/IPoAHalfBridge", "r");
         	if (fp) {
	 	       fread(&myipaddr, 4, 1, fp);	  	        
	 	       fread(&hisipaddr, 4, 1, fp);	          	   
	 	       fclose(fp);	        
	 	       IPOA_HALFBRIDGE_GET_WANIP = TRUE;
         	} else {
         		IPOA_HALFBRIDGE_GET_WANIP = FALSE;
         	}
         
         	// If we can not find public IP in the release table, we should distribute the public IP to DHCP client.      
         	if (( (!(lease = find_lease_by_yiaddr(myipaddr)) ||
         	      /* or it expired and we are checking for expired leases */
	 	    	  (lease_expired(lease))) &&
         	
	 	    	  /* and it isn't on the network */
	 	          (!check_ip(myipaddr))  ) ||
	 	          ((lease = find_lease_by_yiaddr(myipaddr))&&
	 	          !memcmp(lease->chaddr,packet.chaddr,16))) {
  //       		  printf("No host use the public IP\n"); 
         		  FIRST_SEND_PUBLIC_IP = TRUE;
         	} 	
    	}

#ifdef IP_BASED_CLIENT_TYPE
	enum DeviceType devicetype;
	unsigned char *classVendor;
    struct dhcp_ctc_client_info stClientInfo;
    
	if(!(classVendor=get_option(oldpacket, DHCP_VENDOR)))
		//default : PC clients....
		devicetype=CTC_Computer;
	else
	{
        memset(&stClientInfo, 0, sizeof(struct dhcp_ctc_client_info));
		parse_CTC_Vendor_Class(oldpacket, classVendor, &stClientInfo);
        devicetype = (enum DeviceType)(stClientInfo.iCategory);
	}
#endif
	
    	// Modified by Mason Yu for Half Bridge
    	// If user choose Half Bridge feature and the DHCP server send offer is the first time 
    	// after WAN interface is up, we should distribute the WAN IP address to DHCP client.       
    	if ( ((server_config.ippt == 1) && (PPP_HALFBRIDGE_GET_WANIP == TRUE) && (FIRST_SEND_PUBLIC_IP == TRUE)) ||
		((server_config.ippt == 1) && (IPOA_HALFBRIDGE_GET_WANIP == TRUE) && (FIRST_SEND_PUBLIC_IP == TRUE))) {           
    	     	/* Put the ip address into packet that WAN get from PPPoX Server  */ 
	     		packet.yiaddr = myipaddr;  
		               
    	} 
	else {	
	         /* ADDME: if static, short circuit */
	         /* the client is in our lease/offered table */
			 u_int32_t ip_addr;
			 ip_addr = find_IP_by_Mac(oldpacket->chaddr);
			 
			 if ((lease = find_lease_by_chaddr(oldpacket->chaddr))
			 	&& (!find_Mac_by_IP(lease->yiaddr))
			 	&& ((ip_addr == 0) || ((ip_addr != 0) && (ip_addr == lease->yiaddr)))
			 	&& (!check_ip(lease->yiaddr))
#ifdef IP_BASED_CLIENT_TYPE
				&& ((lease->yiaddr==myipaddr)||((ip_addr != 0) || (!check_type(lease->yiaddr, devicetype))))
#endif
				// Mason Yu
				&& ( ntohl(lease->yiaddr) >= ntohl(server_config.start) )
				&& ( ntohl(lease->yiaddr) <= ntohl(server_config.end) )
				) {
				if (!lease_expired(lease)) 
	         			lease_time_align = lease->expires - time(0);	         	
	         		packet.yiaddr = lease->yiaddr;	         		
	         	
	        	 /* Or the client has a requested ip */
				// jim: starzhang added it to support ip/mac binding.
	         	} 	         	
	         	else if(packet.yiaddr = find_IP_by_Mac(packet.chaddr))			
	         	{	         		
	         		goto getip;
		   	}
			else if ((req = get_option(oldpacket, DHCP_REQUESTED_IP)) &&
             
		         	   /* Don't look here (ugly hackish thing to do) */
		         	   memcpy(&req_align, req, 4) &&
	             
		         	   /* and the ip is in the lease range */
		         	   ntohl(req_align) >= ntohl(server_config.start) &&
		         	   ntohl(req_align) <= ntohl(server_config.end) && 
		         	   (!find_Mac_by_IP(req_align)) &&
		         	   /*add by star, for static IP based Mac*/
		         	   /* and its not already taken/offered */ /* ADDME: check that its not a static lease */
		         	   ((!(lease = find_lease_by_yiaddr(req_align)) ||
		         	   
		         	   /* or its taken, but expired */ /* ADDME: or maybe in here */
		         	   lease_expired(lease)))
		         	   /* and it isn't on the network */
		    	     	     && !check_ip(req_align)
#ifdef IP_BASED_CLIENT_TYPE
					&& (!check_type(req_align, devicetype))
#endif
	    	     	     ) {	    	     	     		      	   	        
	         			packet.yiaddr = req_align; /* FIXME: oh my, is there a host using this IP? */
        
	         	/* otherwise, find a free IP */ /*ADDME: is it a static lease? */
	        	} 
			else {
				// Mason Yu. Find IP base on Mac
	//			packet.yiaddr = find_IP_by_Mac(packet.chaddr);
		        	
		        	// Modified by Mason Yu
		        	//packet.yiaddr = find_address(0);
	  	
		        	if (!packet.yiaddr){
#ifdef IP_BASED_CLIENT_TYPE
					 packet.yiaddr = find_address(0, devicetype);
#else
		        		 packet.yiaddr = find_address(0);	 
#endif
		         	}
		         	
		         	/* try for an expired lease */
		         	if (!packet.yiaddr){	
#ifdef IP_BASED_CLIENT_TYPE
					packet.yiaddr = find_address(1, devicetype);
#else
		         		packet.yiaddr = find_address(1);
#endif
		         	}	
	         	}
    	}

getip:

	if(!packet.yiaddr) {
		LOG(LOG_WARNING, "no IP addresses to give -- OFFER abandoned");
		return -1;
	}	

	if (!add_lease(packet.chaddr, packet.yiaddr, server_config.offer_time)) {
		LOG(LOG_WARNING, "lease pool is full -- OFFER abandoned");
		return -1;
	}	

	// Modified by Mason Yu for Half Bridge
	// Asigned Lease Time 
	if ( (((server_config.ippt == 1) && (PPP_HALFBRIDGE_GET_WANIP == TRUE) && (FIRST_SEND_PUBLIC_IP == TRUE)) || 
	      ((server_config.ippt == 1) && (PPP_HALFBRIDGE_GET_WANIP == TRUE) && (RENEW_PUBLIC_IP == TRUE))) ||
	     (((server_config.ippt == 1) && (IPOA_HALFBRIDGE_GET_WANIP == TRUE) && (FIRST_SEND_PUBLIC_IP == TRUE)) || 
	      ((server_config.ippt == 1) && (IPOA_HALFBRIDGE_GET_WANIP == TRUE) && (RENEW_PUBLIC_IP == TRUE))) )  {		
		// Asigned Lease Time for Half Bridge
            	//printf("server_config.ipptlt(sendOffer) = %d\n", server_config.ipptlt); 
            	//server_config.ipptlt = 30;            	 	    
            	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(server_config.ipptlt));
            
            	// Added by Mason Yu
            	if (!add_lease(packet.chaddr, packet.yiaddr, server_config.ipptlt)) {
			LOG(LOG_WARNING, "IPPT: lease pool is full -- OFFER abandoned");
			return -1;
	    	}
	    
    	}else {
	       if ((lease_time = get_option(oldpacket, DHCP_LEASE_TIME))) {
	       		memcpy(&lease_time_align, lease_time, 4);
	       		lease_time_align = ntohl(lease_time_align);
	       		if (lease_time_align > server_config.lease) 
	       			lease_time_align = server_config.lease;
	       }          

	       /* Make sure we aren't just using the lease time from the previous offer */
	       if (lease_time_align < server_config.min_lease) 
	       		lease_time_align = server_config.lease;
	       /* ADDME: end of short circuit */		
	       add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));
    	}

	curr = server_config.options;
	// Modified by Mason Yu for Half Bridge 
	// The following two situations, we should (1) set 0xffffffff as submet and
	// (2) set PPPoX server IP address as Gateway.
	// (1) The user choose Half Bridge feature and DHCP server send offer first time.
	// (2) The user choose Half Bridge feature and user renew public IP.	
	if ( (((server_config.ippt == 1) && (PPP_HALFBRIDGE_GET_WANIP == TRUE) && (FIRST_SEND_PUBLIC_IP == TRUE)) || 
	      ((server_config.ippt == 1) && (PPP_HALFBRIDGE_GET_WANIP == TRUE) && (RENEW_PUBLIC_IP == TRUE))) ||
	     (((server_config.ippt == 1) && (IPOA_HALFBRIDGE_GET_WANIP == TRUE) && (FIRST_SEND_PUBLIC_IP == TRUE)) || 
	      ((server_config.ippt == 1) && (IPOA_HALFBRIDGE_GET_WANIP == TRUE) && (RENEW_PUBLIC_IP == TRUE))) )  {
		//printf("Set gw and subnet by Half Bridge(sendoffer)\n");
		while (curr) {
	       		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)  {
		        	//(1) Set 0xffffffff as submet of the option on this offer.
		        	if (curr->data[OPT_CODE] == DHCP_SUBNET) {			        
		                  	add_simple_option(packet.options, DHCP_SUBNET, 0xffffffff );               
		               
		        	// (2) Set WAN interface address as Gateway of the option on this offer.     
		        	}else if ( curr->data[OPT_CODE] == DHCP_ROUTER ) {	 		              
		                  	add_simple_option(packet.options, DHCP_ROUTER, hisipaddr ); 
		               	
		                // Added by Mason Yu. Access internet fail.
		                // (3) Set Dns Server of the option on this offer.     
		        	}else if ( (PPP_HALFBRIDGE_GET_WANIP) && (curr->data[OPT_CODE] == DHCP_DNS_SERVER) ) {	 		              
		                  	add_simple_option(packet.options, DHCP_DNS_SERVER, dnsipaddr );
		                  	            
		        	}else {
					add_option_string(packet.options, curr->data);
		        	}    			    
	        	}		
	        	curr = curr->next;
	      	}
	      
	      	FIRST_SEND_PUBLIC_IP = FALSE;
	      	RENEW_PUBLIC_IP = FALSE;
	      
	} else {
	      while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
	      		add_option_string(packet.options, curr->data);
	      	curr = curr->next;
	      }
    	}
#if defined(CTC_DHCP_OPTION43) 
		// jim should attach option  43 to dhcp ack packet.
	add_option_string(packet.options, option43);
#endif
#if defined(CTC_DHCP_OPTION60)
		// jim should attach option  60 to dhcp ack packet.
	add_option_string(packet.options, option60);
#endif
	add_bootp_options(&packet);
	
	addr.s_addr = packet.yiaddr;
	LOG(LOG_INFO, "sending OFFER of %s", inet_ntoa(addr));
	return send_packet(&packet, 0);
}


int sendNAK(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;

	init_packet(&packet, oldpacket, DHCPNAK);
	
	DEBUG(LOG_INFO, "sending NAK");
	return send_packet(&packet, 1);
}

#ifdef IP_BASED_CLIENT_TYPE
int set_CTC_port_forwarding(int iSet, unsigned long ulIp, unsigned short usProtocol, unsigned short usPort)
{
    char szAction[4];
    char szProtocol[4];
    unsigned char aucPort[16];
    unsigned char aucDstIP[32];        

    switch (usProtocol)
    {
        case PF_UDP:
        case PF_TCP:
            snprintf(szProtocol, sizeof(szProtocol), "%s", PF_UDP == usProtocol ? "udp" : "tcp");
            break;
            
        case PF_TCP_UDP:
            if (set_CTC_port_forwarding(iSet, ulIp, PF_UDP, usPort))
            {
                if (!set_CTC_port_forwarding(iSet, ulIp, PF_TCP, usPort))
                {
                    if (iSet)
                    {
                        /* undo udp setting */
                        set_CTC_port_forwarding(!iSet, ulIp, PF_UDP, usPort);
                    }
                }
                else
                {
                    /* success */
                    return 1;
                }
            }
            else
            {
                if (!iSet)
                {
                    /* Clear */
                    set_CTC_port_forwarding(iSet, ulIp, PF_TCP, usPort);
                }
            }
            return 0;

        default:
            return 0;
    }

    snprintf(szAction, sizeof(szAction), "%s", iSet ? "-A" : "-D");
    snprintf(aucPort, sizeof(aucPort), "%u", usPort);
	snprintf(aucDstIP, sizeof(aucDstIP), "%s:%s", inet_ntoa(*((struct in_addr *)(&ulIp))), aucPort);

	if (0 == va_cmd(IPTABLES, 15, 1, "-t", "nat", szAction, "PREROUTING", "-i", "!", 
                    server_config.interface, "-p", szProtocol, "--dport",
                    aucPort, "-j", "DNAT", "--to-destination", aucDstIP))
	{
        /* Set ip filter */
        snprintf(szAction, sizeof(szAction), "%s", iSet ? "-I" : "-D");
    	if (0 != va_cmd(IPTABLES, 13, 1, szAction, "ipfilter", "-i", "!", 
                        server_config.interface, "-o", server_config.interface,                         
                        "-p", szProtocol, "--dport", aucPort, "-j", "RETURN"))
    	{
            if (iSet)
            {
                /* undo nat setting */
                va_cmd (IPTABLES, 15, 1, "-t", "nat", "-D", "PREROUTING", "-i", "!", 
                        server_config.interface, "-p", szProtocol, "--dport",
                        aucPort, "-j", "DNAT", "--to-destination", aucDstIP);
            }
    	}
        else
        {
            /* success */
            return 1;
        }
	}
    else
    {
        if (!iSet)
        {
            /* go on clear ipfilter */
            va_cmd (IPTABLES, 13, 1, szAction, "ipfilter", "-i", "!", 
                    server_config.interface, "-o", server_config.interface,                         
                    "-p", szProtocol, "--dport", aucPort, "-j", "RETURN");
        }
    }

    return 0;
}

void CTC_lease_expired(struct dhcpOfferedAddr *pstLease)
{
    struct dhcp_ctc_client_info *pstClientInfo;

    pstClientInfo = (struct dhcp_ctc_client_info *)(pstLease->stClientInfo.pvDHCPClientData);
    if ((0 != pstClientInfo->stPortForwarding.usPort)
        && (pstClientInfo->stPortForwarding.iSet))
    {
        set_CTC_port_forwarding(0, pstLease->yiaddr, 
                                 pstClientInfo->stPortForwarding.usProtocol, 
                                 pstClientInfo->stPortForwarding.usPort);        
    }

    free(pstClientInfo);

    return;
}
#endif

int sendACK(struct dhcpMessage *oldpacket, u_int32_t yiaddr)
{
	struct dhcpMessage packet;
	struct option_set *curr;
	unsigned char *lease_time;
	u_int32_t lease_time_align = server_config.lease;
	struct in_addr addr;
    	// Added by Mason Yu for Half Bridge
    	// Added by Mason Yu. Access internet fail.
	unsigned long myipaddr, hisipaddr, dnsipaddr;  
	FILE *fp;
	struct dhcpOfferedAddr *pstLease;
	myipaddr = 0;
    
	// Added by Mason Yu for Half Bridge	
    if ( server_config.ippt )  {    	 
    	  
    	 fp = fopen ("/tmp/PPPHalfBridge", "r");
         if (fp) {
	        fread(&myipaddr, 4, 1, fp);	         
	        fread(&hisipaddr, 4, 1, fp);
	        // Added by Mason Yu. Access internet fail.
	 	fread(&dnsipaddr, 4, 1, fp);	         
	        
	        fclose(fp);
         } else {
         	PPP_HALFBRIDGE_GET_WANIP = FALSE;
         } 
         
         
         fp = fopen ("/tmp/IPoAHalfBridge", "r");
         if (fp) {
	        fread(&myipaddr, 4, 1, fp);	         
	        fread(&hisipaddr, 4, 1, fp);
	       	         
	        fclose(fp);
         } else {
         	 IPOA_HALFBRIDGE_GET_WANIP = FALSE;
         }
          
         
         if ( yiaddr == myipaddr )  {    	     
	         RENEW_PUBLIC_IP = TRUE;
	// Added by Mason Yu. Access internet fail.
        } else if ( !((ntohl(server_config.start) <= yiaddr) && (yiaddr<= ntohl(server_config.end))) ) {
        	printf("sendACK: This ip is not on DHCP Pool for IP Passthrough\n");         	
        	sendNAK(oldpacket);
         	return -1;
	}
	#if 0
	// Kaohj
	// ippt and bridged IP not yet leased
	else if (myipaddr != 0 && !(pstLease = find_lease_by_yiaddr(myipaddr))) {
		// reject it and let this client be able to discover the bridged IP
        	sendNAK(oldpacket);
         	return -1;
	}
	#endif
	goto init_ack;
	           
    }	
	
	// Added by Mason Yu
	if ( !((ntohl(server_config.start) <= yiaddr) && (yiaddr<= ntohl(server_config.end))) ) {
		printf("sendACK: This ip is not on DHCP Pool\n");         	
        	sendNAK(oldpacket);
         	return -1;
	}
		
init_ack:
	
	init_packet(&packet, oldpacket, DHCPACK);
	packet.yiaddr = yiaddr;
	
	// Modified by Mason Yu for Half Bridge 
	// Asigned Lease Time	
	if ( (((server_config.ippt == 1) && (PPP_HALFBRIDGE_GET_WANIP == TRUE) && (FIRST_SEND_ACK == TRUE)) ||
	      ((server_config.ippt == 1) && (PPP_HALFBRIDGE_GET_WANIP == TRUE) && (RENEW_PUBLIC_IP == TRUE))) ||
	     (((server_config.ippt == 1) && (IPOA_HALFBRIDGE_GET_WANIP == TRUE) && (FIRST_SEND_ACK == TRUE)) ||
	      ((server_config.ippt == 1) && (IPOA_HALFBRIDGE_GET_WANIP == TRUE) && (RENEW_PUBLIC_IP == TRUE))) )  {
            // Asigned Lease Time for Half Bridge
            //printf("server_config.ipptlt(sendAck) = %d\n", server_config.ipptlt);
            add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(server_config.ipptlt));
            lease_time_align = server_config.ipptlt;
    }else {
	        if ((lease_time = get_option(oldpacket, DHCP_LEASE_TIME))) {
	       	     memcpy(&lease_time_align, lease_time, 4);
	       	     lease_time_align = ntohl(lease_time_align);
	       	     if (lease_time_align > server_config.lease) 
	       		     lease_time_align = server_config.lease;
	       	     else if (lease_time_align < server_config.min_lease) 
	       		     lease_time_align = server_config.lease;
	       }
	
	       add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));
	}
	
	curr = server_config.options;
	// Modified by Mason Yu for Half Bridge 	
	if ( (((server_config.ippt == 1) && (PPP_HALFBRIDGE_GET_WANIP == TRUE) && (FIRST_SEND_ACK == TRUE)) ||
	      ((server_config.ippt == 1) && (PPP_HALFBRIDGE_GET_WANIP == TRUE) && (RENEW_PUBLIC_IP == TRUE))) ||
	     (((server_config.ippt == 1) && (IPOA_HALFBRIDGE_GET_WANIP == TRUE) && (FIRST_SEND_ACK == TRUE)) ||
	      ((server_config.ippt == 1) && (IPOA_HALFBRIDGE_GET_WANIP == TRUE) && (RENEW_PUBLIC_IP == TRUE))) )  {
		  
		  while (curr) {
	        if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)  {
		        //(1) Set 0xffffffff as submet of the option on this offer.
		        if (curr->data[OPT_CODE] == DHCP_SUBNET) {		       
		                  add_simple_option(packet.options, DHCP_SUBNET, 0xffffffff );			                   
		                  
		        // (2) Set WAN interface address as Gateway of the option on this offer.     
		        } else if ( curr->data[OPT_CODE] == DHCP_ROUTER ) {			      
		                   add_simple_option(packet.options, DHCP_ROUTER, hisipaddr );
		                   
		        // Added by Mason Yu. Access internet fail.
		        // (3) Set Dns Server of the option on this offer.     
		        }else if ( (PPP_HALFBRIDGE_GET_WANIP) && (curr->data[OPT_CODE] == DHCP_DNS_SERVER) ) {	 		              
		                  add_simple_option(packet.options, DHCP_DNS_SERVER, dnsipaddr );
		                 	               
		        }else {
			               add_option_string(packet.options, curr->data);
		        }    			    
	        }		
	        curr = curr->next;
	      }
	      
	      FIRST_SEND_ACK = FALSE;
	      RENEW_PUBLIC_IP = FALSE;
	       
	} else  {
	      while (curr) {
	      	if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
	      		add_option_string(packet.options, curr->data);
	      	curr = curr->next;
	      }
    }
#if defined(CTC_DHCP_OPTION43) 
		// jim should attach option  43 to dhcp ack packet.
	add_option_string(packet.options, option43);
#endif
#if defined(CTC_DHCP_OPTION60)
		// jim should attach option  60 to dhcp ack packet.
	add_option_string(packet.options, option60);
#endif
	add_bootp_options(&packet);

	addr.s_addr = packet.yiaddr;
	LOG(LOG_INFO, "sending ACK to %s", inet_ntoa(addr));

	if (send_packet(&packet, 0) < 0) 
		return -1;

	pstLease = add_lease(packet.chaddr, packet.yiaddr, lease_time_align);

    if (NULL != pstLease)
    {
        pstLease->stClientInfo.ulDevice = FP_Computer;
    
#ifdef IP_BASED_CLIENT_TYPE
        struct dhcp_ctc_client_info *pstClientInfo;
		unsigned char *classVendor;
        
		if (NULL != (classVendor=get_option(oldpacket, DHCP_VENDOR)))
		{
            pstClientInfo = malloc(sizeof(struct dhcp_ctc_client_info));
            if (NULL != pstClientInfo)
            {
                memset(pstClientInfo, 0, sizeof(struct dhcp_ctc_client_info));
                parse_CTC_Vendor_Class(oldpacket, classVendor, pstClientInfo);
                
                pstLease->stClientInfo.pfLeaseExpired = CTC_lease_expired;
                pstLease->stClientInfo.pvDHCPClientData = (void *)pstClientInfo;

                if ((0 != pstClientInfo->stPortForwarding.usPort)
                    && (!pstClientInfo->stPortForwarding.iSet))
                {
                    if (set_CTC_port_forwarding(1, pstLease->yiaddr, 
                                                 pstClientInfo->stPortForwarding.usProtocol, 
                                                 pstClientInfo->stPortForwarding.usPort))
                    {
                        pstClientInfo->stPortForwarding.iSet = 1;
                    }
                }

                switch (pstClientInfo->iCategory)
                {
                    case CTC_STB:
                        pstLease->stClientInfo.ulDevice = FP_STB;
                        break;

                    case CTC_PHONE:
                        pstLease->stClientInfo.ulDevice = FP_MOBILE;
                        break;

                    default:
                        break;
                }
            }
		}

#ifdef CONFIG_CTC_E8_CLIENT_LIMIT
        update_client_limit(pstLease);
#endif
#endif

        /* update_force_portal(); */
        update_force_portal(pstLease);
    }

#ifdef _CWMP_TR069_
	handle_tr111( oldpacket );
#endif

#if 0
#ifdef CONFIG_CTC_E8_CLIENT_LIMIT
	handle_clients_limit(oldpacket,yiaddr);
#endif
#endif
	return 0;
}


int send_inform(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;
	struct option_set *curr;

	init_packet(&packet, oldpacket, DHCPACK);
	
	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}
#if defined(CTC_DHCP_OPTION43) 
		// jim should attach option  43 to dhcp ack packet.
	add_option_string(packet.options, option43);
#endif
#if defined(CTC_DHCP_OPTION60)
		// jim should attach option  60 to dhcp ack packet.
	add_option_string(packet.options, option60);
#endif
	add_bootp_options(&packet);

#ifdef _CWMP_TR069_
	handle_tr111( oldpacket );
#endif

	return send_packet(&packet, 0);
}



